/***************
 * NAME:            "IconSplit" - version 1.0.0
 * DESCRIPTION:     Functions to split icon images
 * AUTHOR:          (c) Mike Riddle - initial release 03-06-23                     
 ***************/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <dos/dos.h>
#include <exec/exec.h>

#include <proto/wb.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/icon.h>
#include <proto/workbench.h>
#include <proto/intuition.h>

#include <workbench/icon.h>
#include <workbench/workbench.h>
#include <intuition/intuition.h>



/************************************/
/*    MACROS BASED ON MODIFYICON    */
/************************************/

#define MAXICONSIZE 256
#define MAXDEPTH 8
#define MAXCOLORS (1<<MAXDEPTH)


/* Defines for Get Data Functions - based on modifyicon */
#define GetUint32(a)  (((((unsigned char *)a)[0])<<24)|((((unsigned char *)a)[1])<<16)| \
                       ((((unsigned char *)a)[2])<< 8)|((((unsigned char *)a)[3])))
#define GetUint16(a)  (((((unsigned char *)a)[0])<< 8)|((((unsigned char *)a)[1])))

/* Defines for Put Data Functions - based on modifyicon */
#define PutUint32(a, b) {unsigned long epu32 = (b); (a)[0] = (unsigned char) (epu32 >> 24); \
                         (a)[1] = (unsigned char) (epu32 >> 16); \
                         (a)[2] = (unsigned char) (epu32 >> 8); \
                         (a)[3] = (unsigned char) epu32;}
#define PutUint16(a, b) {unsigned short epu16 = (b); (a)[0] = (unsigned char) (epu16 >> 8); \
                         (a)[1] = (unsigned char) epu16;}

#define PutUint32Inc(a, b) {PutUint32(a,b); (a) += 4;}
#define PutUint16Inc(a, b) {PutUint16(a,b); (a) += 2;}


enum DecodeType { DECODE_PALETTE, DECODE_IMAGE };

STRPTR srcename = NULL;

/****************************************************************************************/

/*************************************/
/*    STRUCTS BASED ON MODIFYICON    */
/*************************************/

struct IconPalette 
{
  unsigned char      Red;
  unsigned char      Green;
  unsigned char      Blue;
};

struct IconImage 
{
  LONG               Width;
  LONG               Height;
  LONG               Depth;
  LONG               NumColors;
  LONG               Transparency; /* Transparent Color or -1 */  
  unsigned char      ImageData[MAXICONSIZE*MAXICONSIZE*MAXDEPTH/8]; /* Chunky Data */  
  struct IconPalette Palette[MAXCOLORS];  
};

struct IconImageCollection 
{
  struct IconImage NewImage[2];
};

/****************************************************************************************/

LONG GetFileSize(char *imName);
LONG Get_IFFOffset(char *srcename); //ReadIcon_IFF
static char *ReadIcon(char *srcename);
char *Check_Icon_Signature(char *srcename); //ReadIcon
BOOL ReadIconImage_IFF(char *srcename, LONG iffoffset);
static BOOL SaveImage(char *name, struct IconImage *im);
static void WriteData(char *filename, unsigned char *data, LONG datasize);
//static long EncodeRLE(unsigned char *srce, unsigned char *dest, int len);
static void SetPalette(struct IconImage *im, const struct IconPalette *pal, long num); 
static int ChunkyToBitplanes(UBYTE *chunkyBuffer, UBYTE *bitBuffer, LONG bitPlanes, LONG imagewidth);
static void MakePlanar(unsigned char *inbuffer, unsigned char *outbuffer, LONG width, LONG height, LONG depth);
static char *createrle(unsigned long depth, unsigned char *dtype, LONG *dsize, unsigned long size, unsigned char *src);
static char *EncodeRLE(unsigned long depth, unsigned char *dtype, LONG *dsize, unsigned long size, unsigned char *src);
static char *DecodeNew(unsigned char *buffer, long numbytes, long bits, struct IconImage *im, enum DecodeType type, long mode);
static char *DecodeRLE(unsigned char *srce, unsigned long bytes, unsigned long len, struct IconImage *im);
static char *DecodeBytes(unsigned char *srce, unsigned long bytes, unsigned long depth, struct IconImage *im, unsigned long len);

/**************************************************************************************************/

LONG GetFileSize(char *imName)
{
    BPTR file = Open(imName, MODE_OLDFILE);
    struct FileInfoBlock *fib = (struct FileInfoBlock*)AllocDosObject( DOS_FIB, NULL );
    LONG size = -1;
    
    if (fib)
    {
	    if (ExamineFH( file, fib ))
	    {
	        if ((fib->fib_DirEntryType > 0) && (fib->fib_DirEntryType != ST_SOFTLINK))
	        {
		        size = 0;
	        }
	        size = fib->fib_Size;
	    }
	    FreeDosObject( DOS_FIB, fib );
    }
    Close(file);
    
    return size;
}

/**************************************************************************************************/

//See: static char *ReadIcon(char *srcename)
char *Check_Icon_Signature(char *srcename)
{	
	char *icontype;
	UBYTE fileID[5];
    unsigned char buffer[4];

	Printf ("Icon Filename: %s\n", srcename);    
	
	if(srcename) 
	{
        BPTR file = Open(srcename, MODE_OLDFILE);
        if(file)
		{
            Seek(file,0,OFFSET_BEGINNING);
	        Read(file,fileID, 4); //Read(file,buffer, 4);
	        fileID[4] = ('\0');

			//if(GetUint32(buffer) == 0x89504E47)
            if( fileID[1] == 'P' && fileID[2] == 'N' && fileID[3] == 'G' )
            {				
                icontype = "png";				           
            }
            //else if(GetUint16(buffer) == 0xE310)
			else if( fileID[0] == 0xE3 && fileID[1] == 0x10 )
            {                           
                icontype = "iff";                
            }
		}
		Close(file);
	}    

    return icontype;
}

/**************************************************************************************************/

static char *ReadIcon(char *srcename)
{
    char *icontype;
    unsigned char buffer[4];

    if(srcename)
    {
        BPTR file = Open(srcename, MODE_OLDFILE);
        if(file)
        {
            Read(file,buffer, 4);
            if(GetUint32(buffer) == 0x89504E47)
                icontype = "png";
            else if(GetUint16(buffer) == 0xE310)
                icontype = "iff";
            else
                icontype = "none";
            Close(file);
        }
    }

    return icontype;
} 

/**************************************************************************************************/

static void WriteData(char *new_filename, unsigned char *data, LONG datasize)
{  
    BPTR fh = Open( new_filename, MODE_NEWFILE );
	if(fh)
	{        
	    Seek(fh,0,OFFSET_BEGINNING);
        Write(fh, data, datasize);
	}	
    Close( fh );
}

/**************************************************************************************************/

LONG Get_IFFOffset(char *srcename)
{	
	LONG offset;
	//LONG imSize;
	LONG i, j, k;	
	UWORD FileID;		
	UBYTE Buffer16[2];
	UBYTE Buffer32[4];	
	WORD width, height, depth;
	int imagesize, imageoffset;

	//Printf("\nGet_IFFOffset...\n");

	//unsigned char *buffer;
	LONG firstimage, secondimage;	
	LONG tooltypes, deftool, drawerdata, toolwin;	

	/* Get total icon file size */
	//imSize = GetFileSize(srcename);
	//Printf ("File Size: %ld\n", imSize);

	/* Get byte array of IFF data */
    //buffer = (unsigned char *)AllocVec(imSize, MEMF_ANY);

	/* Open icon file to Read */
    BPTR fh = Open(srcename, MODE_OLDFILE);
    if (fh)
	{
		offset = 0;

		//Seek(fh,offset,OFFSET_BEGINNING);
		//Read(fh, buffer, imSize);        		  

        Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer16, 2);	    		
		if( Buffer16[0] == 0xE3 && Buffer16[1] == 0x10 ) 
		//Printf ("Found Classic Icon File!\n");

		offset += 2;
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer16, 2);		
		FileID = GetUint16(Buffer16);
		if(FileID == 1)
		//Printf("Found Valid Version Number!\n");

		offset += 20; //22
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer32, 4);
		//secondimage = EndGetM32(buffer+0x16) ? 1 : 0;		
		//firstimage = GetUint32(Buffer32) ? 1: 0; //22
		firstimage = (GetUint32(Buffer32)==0) ? 0: 1; //22
		//Printf ("First Image: %ld, %ld\n", firstimage, offset);

		offset += 4; //26
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer32, 4);
		//secondimage = GetUint32(Buffer32) ? 1: 0; //26
		secondimage = (GetUint32(Buffer32)==0) ? 0: 1; //26		
		//Printf ("Second Image: %ld, %ld\n", secondimage, offset);

		offset += 24; //26+24=50
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer32, 4);
		//deftool = GetUint32(Buffer32) ? 1: 0; //50
		deftool = (GetUint32(Buffer32)==0) ? 0: 1; //50
		//Printf ("DiskObject - deftool: %ld, %ld\n", deftool, offset);
		offset += 4; //54
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer32, 4);
		//tooltypes = GetUint32(Buffer32) ? 1: 0; //54
		tooltypes = (GetUint32(Buffer32)==0) ? 0: 1; //54
        //Printf ("DiskObject - tooltypes: %ld, %ld\n", tooltypes, offset);
		offset += 12; //66
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer32, 4);
        //drawerdata = GetUint32(Buffer32) ? 1: 0; //66
		drawerdata = (GetUint32(Buffer32)==0) ? 0: 1; //66
		//Printf ("DiskObject - drawerdata: %ld, %ld\n", drawerdata, offset);        
		offset += 4; //70
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer32, 4);
		//toolwin = GetUint32(Buffer32) ? 1: 0; //70
		toolwin = (GetUint32(Buffer32)==0) ? 0: 1; //70
		//Printf ("DiskObject - toolwin: %ld, %ld\n", toolwin, offset);
		offset += 8; //64+14=78
		//Printf("Current Offset: %ld\n", offset); //78 ??

		if(drawerdata == 1)
        {		    
		    offset += 56; //78+56=134
		}
		if(firstimage == 1)
		{
			offset += 4;
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer16, 2);
            width = GetUint16(Buffer16); //offset + 4			           
			//Printf ("First Image Width: %ld, 0x%x%x, %ld\n", width, Buffer16[0],Buffer16[1], offset);

			offset += 2;
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer16, 2);
			height = GetUint16(Buffer16); //offset + 6			
			//Printf ("First Image Height: %ld, 0x%x%x, %ld\n", height, Buffer16[0],Buffer16[1], offset);

			offset += 2;
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer16, 2);
			depth = GetUint16(Buffer16); //offset + 8			
			//Printf ("First Image Depth: %ld, 0x%x%x, %ld\n", depth, Buffer16[0],Buffer16[1], offset);

			offset += 12; //offset + 20
			imagesize = (((width+15)&(~15)) * height);
			//imageoffset = (((((width+15)&(~15)) * height) * depth)>>3);
			imageoffset = ((imagesize * depth)>>3);

			//Printf ("First Image Size: %ld, %ld\n", imageoffset, offset);
			offset += imageoffset;
		}
		if(secondimage == 1)
		{
			offset += 4;
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer16, 2);
            width = GetUint16(Buffer16); //offset + 4			           
			//Printf ("Second Image Width: %ld, 0x%x%x, %ld\n", width, Buffer16[0],Buffer16[1], offset);

			offset += 2;
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer16, 2);
			height = GetUint16(Buffer16); //offset + 6			
			//Printf ("Second Image Height: %ld, 0x%x%x, %ld\n", height, Buffer16[0],Buffer16[1], offset);

			offset += 2;
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer16, 2);
			depth = GetUint16(Buffer16); //offset + 8			
			//Printf ("Second Image Depth: %ld, 0x%x%x, %ld\n", depth, Buffer16[0],Buffer16[1], offset);
			
			offset += 12; //offset + 20
			imagesize = (((width+15)&(~15)) * height);
			//imageoffset = (((((width+15)&(~15)) * height) * depth)>>3);
			imageoffset = ((imagesize * depth)>>3);						

			//Printf ("Second Image Size: %ld, %ld\n", imageoffset, offset);
			offset += imageoffset;
		}
		if(deftool == 1)
        {
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer32, 4);
			i = GetUint32(Buffer32);            
			offset += i+4;
			//Printf("Current Offset: %ld\n", offset);
		}
		if(tooltypes == 1)
		{
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer32, 4);
			i = GetUint32(Buffer32)/4-1;
			offset += 4;
			
			for(j = 0; j < i; ++j)
            {
				Seek(fh,offset,OFFSET_BEGINNING);
		        Read(fh,Buffer32, 4);
                k = GetUint32(Buffer32);
				offset += k+4;
            }
			//Printf("Current Offset: %ld\n", offset);
		}
		if(toolwin == 1)
        {
			Seek(fh,offset,OFFSET_BEGINNING);
		    Read(fh,Buffer32, 4);
            i = GetUint32(Buffer32);        
		    offset += i+4; 
			//Printf("Current Offset: %ld\n", offset);
        }
		/* Check for Beginning of OS3.5 IFF Data */
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh,Buffer32, 4); //0x464F524D) = 'FORM'		
		if( Buffer32[0] == 0x46 && Buffer32[1] == 0x4F && Buffer32[2] == 0x52 && Buffer32[3] == 0x4D)
		{		
		    //Printf("Found Beginning of OS3.5 IFF Data: %ld\n", offset);
			return offset;
		}
		else
		{
			/* Check Again! */
			offset += 6; //Skip 6 zero bytes at end
			Seek(fh,offset,OFFSET_BEGINNING);
			Read(fh,Buffer32, 4);
			if( Buffer32[0] == 0x46 && Buffer32[1] == 0x4F && Buffer32[2] == 0x52 && Buffer32[3] == 0x4D)
		    {		
		        //Printf("Found Beginning of OS3.5 IFF Data: %ld\n", offset);
				return offset;
		    }
		}
	}
	Close(fh);	
    //Printf("Current Offset: %ld\n", offset);

	return offset;
}

/**************************************************************************************************/


/****************************************************************************/
/*                         SAVE ICON IMAGE FUNCTIONS                        */
/****************************************************************************/


/**************************************************************************************************/

static int ChunkyToBitplanes(UBYTE *chunkyBuffer, UBYTE *bitBuffer, LONG bitPlanes, LONG imagewidth) 
{    
    int i, j;
    int hOffset = 0; //scanline offsets
    int bitplaneIndex = 0; //Bitplane index
    //int alignedwidth = (imagewidth + 15) & ~15; //Modulus 16   
    int bytesPerLine = (imagewidth / 8); //Bitplane length
    int rowSize = (bytesPerLine * bitPlanes); //Planar scanline length

    /* Converts one scanline from chunky to planar */

    //Printf("ChunkyBytesToBitplanes");
    
    //Block of 8 bytes    
    UBYTE colorBytes[8];   
    
        
    while (hOffset < bytesPerLine)
    {
        //Copy 8 bytes to new buffer. Chunky Buffer is Modulus 16 & therefore Modulus 8.
        memcpy(colorBytes, chunkyBuffer + (hOffset * 8), 8);       
        for (j = 0; j < bitPlanes; j++) //Fill bitplane byte locations (Up to bytesPerLine)
        {
            bitplaneIndex = (j * bytesPerLine);
            //UBYTE bitmask = (UBYTE)(1 << bit);
            
            for (i = 0; i < 8; i++) //Set corresponding bits in bitplane buffers.
            {
                /* Check if the current bit of the index value is set */
                if ((colorBytes[i] & (1 << j)) != 0)
                {                    
                    /* (Bit is Set (1)) Modify the current bit in the bitplane byte to be 1 and leave the others untouched */
                    bitBuffer[bitplaneIndex + hOffset] |= (UBYTE)(1 << 7 - i);
                    //Set bit = number |= 1 << x; //return value |= (UBYTE)(1 << pos);
                }
                else
                {                    
                    /* (Bit is Not Set (0)) Modify the current bit in the bitplane byte to be 0 and leave the others untouched */
                    bitBuffer[bitplaneIndex + hOffset] &= (UBYTE)(~(1 << 7 - i));
                    //Clear bit = number &= ~(1 << x); //return value &= (UBYTE)(~(1 << pos));
                }
            }            
        }
        hOffset++;
	
    } //End while loop

    return 0;        
}

/**************************************************************************************************/

static void MakePlanar(unsigned char *inbuffer, unsigned char *outbuffer, LONG width, LONG height, LONG depth)
{
    int y; //Y-Loop iterator
    LONG kOffset, pOffset; //chunky & planar scanline offsets   
    LONG alignedwidth = (width + 15) & ~15; //Chunky scanline length //Modulus 16
    LONG bytesPerRow = (alignedwidth / 8); //Bitplane length
    LONG linelength = (bytesPerRow * depth); //Planar scanline length

    //Printf("MakePlanar");

    /* Set up line buffer to receive chunky data */
    UBYTE chunkyBuffer[alignedwidth];

    /* Set up line buffer to receive planar data */
    UBYTE bitBuffer[linelength];

    /* Set up Y-Loop to process all scanlines */
    for(y = 0; y < height; y++)
    {
        kOffset = (width * y); /* chunky offset */
        memcpy(chunkyBuffer, inbuffer+kOffset, width); /* copy chunky data */       

        /* Convert Chunky ScanLine to Planar ScanLine */	      
        ChunkyToBitplanes(chunkyBuffer, bitBuffer, depth, alignedwidth);

        pOffset = (linelength * y); /* planar offset */
        memcpy(outbuffer+pOffset, bitBuffer, linelength); /* copy planar data */        
    }
}

/**************************************************************************************************/

static BOOL SaveImage(char *name, struct IconImage *im)
{     
    /* Save Image to ILBM File */

    int i;
    LONG transp;
    LONG numpal = 0;    
    LONG ilbmsize = 0;
	LONG formsize = 0;
    LONG numcolors = 0;
    LONG width, height, depth;

    unsigned char *imdata;
    unsigned char *imgbits;    

    width = im->Width;
    height = im->Height;
    depth = im->Depth; //8;
    transp = im->Transparency;
    numcolors = 1<<depth; //256 //No. colors by depth
	numpal = numcolors*3; //Colormap size in bytes     
	     

    UBYTE rgb[numcolors*3];
    for(i = 0; i < im->NumColors; ++i)
    {
        //Convert Palette to CMAP Byte Array
        rgb[i * 3] = (UBYTE)im->Palette[i].Red;
        rgb[(i * 3)+1] = (UBYTE)im->Palette[i].Green;
        rgb[(i * 3)+2] = (UBYTE)im->Palette[i].Blue;
    }
    
    LONG alignedwidth = (im->Width + 15) & ~15; //Modulus 16
    LONG bytesPerRow = (alignedwidth / 8); 
    LONG linelength = (bytesPerRow * depth); //Planar scanline
    LONG planarsize = (linelength * im->Height);

    Printf ("depth: %ld\n", depth);
    Printf ("alignedwidth: %ld\n", alignedwidth);
	Printf ("bytesPerRow: %ld\n", bytesPerRow);
	Printf ("linelength: %ld\n", linelength);
	Printf ("planarsize: %ld\n", planarsize);
        
    imgbits = AllocMem( planarsize, MEMF_ANY );    

    MakePlanar(im->ImageData, imgbits, width, height, depth);
        
    
    ilbmsize = ( 40 + (numpal+8) + (planarsize+8) );
    formsize = (ilbmsize - 8);

    Printf ("ILBM Planar File Size: %ld\n", ( 40 + (numpal+8) + (planarsize+8) ));
	Printf ("ILBM Planar Form Size: %ld\n", ( 40 + (numpal+8) + (planarsize) ));
    
    imdata = AllocMem( ilbmsize, MEMF_ANY );    
    

    /* Build new byte array for bmhd */
    unsigned char bmhd[20];
    bmhd[0] =  (width >> 8); //Width
    bmhd[1] =  (width & 0xFF);
    bmhd[2] =  (height >> 8); //Height
    bmhd[3] =  (height & 0xFF);
	bmhd[4] =  bmhd[5] = 0; //Left Offset
    bmhd[6] =  bmhd[7] = 0; //Top Offset
    bmhd[8] =  depth;
	bmhd[9] =  0; //Masking
    bmhd[10] = 0; //Compression //imgfmt
    bmhd[11] = 0; //Padding
    bmhd[12] = (transp>> 8); 
    bmhd[13] = (transp & 0xFF);
    bmhd[14] = 10; //XAspect;
    bmhd[15] = 11; //YAspect;
    bmhd[16] = (width >> 8); //PageWidth
    bmhd[17] = (width & 0xFF);
    bmhd[18] = (height >> 8); //PageHeight
    bmhd[19] = (height & 0xFF);

    /* Build new byte array with ilbm data */
    PutUint32(imdata, 0x464F524D); //ID_FORM 0x464F524D					
	PutUint32(imdata+4, formsize); 
	PutUint32(imdata+8, 0x494C424D); //ID_ILBM 0x494C424D
	PutUint32(imdata+12, 0x424D4844); //ID_BMHD 0x424D4844
	PutUint32(imdata+16, 20);
	memcpy(imdata+20, bmhd, 20);
    PutUint32(imdata+20+20, 0x434D4150); //ID_CMAP 0x434D4150
	PutUint32(imdata+24+20, numpal);
    memcpy(imdata+28+20, rgb, numpal);
    PutUint32(imdata+48+numpal, 0x424F4459); //ID_BODY 0x424F4459     
    PutUint32(imdata+52+numpal, planarsize);  
    memcpy(imdata+56+numpal, imgbits, planarsize);
    
    /* Write ILBM */
    WriteData(name, imdata, ilbmsize);

    /* Free Resources */
    if (imdata) FreeMem( imdata, ilbmsize );
    if (imgbits) FreeMem( imgbits, planarsize );    
    
    return TRUE;
}

/**************************************************************************************************/

static void SetPalette(struct IconImage *im, const struct IconPalette *pal, long num)
{

  /**************************************/
  /*    FUNCTION BASED ON MODIFYICON    */
  /**************************************/  

  long i, j;

  for(j = 0; j < im->NumColors; ++j)
  {
    i = j%num;
    im->Palette[j].Red = pal[i].Red;
    im->Palette[j].Green = pal[i].Green;
    im->Palette[j].Blue = pal[i].Blue;
  }
}

/**************************************************************************************************/

static char *EncodeRLE(unsigned long depth, unsigned char *dtype, LONG *dsize, unsigned long len, unsigned char *srce)
{
    int i;
    unsigned char val;
    long idx, pdx, cnt;    
    unsigned char *dest;
    unsigned char c1, c2;         
    long out, count, equal;
    //unsigned long bitbuf, numbits; 
    
    out = 0;
    count = 0;
    equal = 1;
    //depth = 5;
    //numbits = 0;

    dest = malloc(len * 2);
    if (!dest) return NULL;

    Printf ("EncodeRLE - Number of Bits: %ld\n", depth); 

    /* byte arrays to test rle encoding */

    unsigned char input[] = {0xAA, 0xAA, 0xAA, 0x80, 0x00, 0x2A, 0xAA, 0xAA, 0xAA, 0xAA, 0x80, 0x00, 0x2A, 0x22,
    0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA };
    // original = AA AA AA 80 00 2A AA AA AA AA 80 00 2A 22 AA AA AA AA AA AA AA AA AA AA
    // output is: 'FE AA 02 80 00 2A FD AA 03 80 00 2A 22 F7 AA'

    //unsigned char[] input = {AA AA AA C6 C6 C6 E3 E3 E3 FF FF FF 8A 8A 8A AA AA AA 6A 6A 6A 29 29 29 49 49 49 56 B1 5A 00 00 00 19 5E 32 38 88 46 24 42 66 52 66 94 7F 89 C2 AD AD F0};
    //FE AA FE C6 FE E3 FE FF FE 8A FE AA FE 6A FE 29 FE 49 02 56 B1 5A FE 00 11 19 5E 32 38 88 46 24 42 66 52 66 94 7F 89 C2 AD AD F0

    //unsigned char[] input = {AA AA AA DF BA 45 EF E7 14 FF FF FF AA AA AA C6 C6 C6 E3 E3 E3 6A 6A 6A 8A 8A 8A 49 49 49 00 00 00 29 29 29 38 88 46 19 5E 32 24 42 66 52 66 94 7F 89 C2};
    //FE AA 05 DF BA 45 EF E7 14 FE FF FE AA FE C6 FE E3 FE 6A FE 8A FE 49 FE 00 FE 29 0E 38 88 46 19 5E 32 24 42 66 52 66 94 7F 89 C2

    c1 = srce[0]; /* previous byte */
    //c1 = input[0]; /* previous byte */
    for (idx = 1; idx < len+1; idx++)
    {
        c2 = srce[idx]; /* next byte */
        //c2 = input[idx]; /* next byte */
        if ((idx < len) && (c1 == c2)) /* There is a byte run - value is same */
	    {            
		    Printf ("EncodeRLE - Bytes are equal!! = %ld\n", equal);
            equal++; /* increment equal */

            /* Force 'if (runCount == 128)' because we start
			at runCount = 1 so that we have all 127 bytes */
		    if (equal == 128) //if (runCount == 128)
		    {				
				/* Don't exceed 127 max bytes */		
				equal = 128;          		          
			    dest[out] = 256 - (equal - 1);		                			              
                dest[out+1] = c1;		               		                
		        cnt = 2; //2-byte pair

                val = 256 - (equal - 1);
                Printf ("EncodeRLE - RLE actual value: %d\n", val);
                Printf ("EncodeRLE - RLE loop value: %d\n", equal);
                Printf ("EncodeRLE - RLE whole byte: %d\n", c1);				            
                    				
			    /* reset run */
                out += cnt;
                equal = 1;
	    }
	    else if (idx < len && equal*depth <= 16) //2 not equal is 2 (or less)*8 <=16
	    {
	        /* Runs of two bytes adjacent to non-runs are typically written as literal data */
	        //count += equal; //increment count to add to adjacent non-run.
	        //equal = 1; //reset equal to 1
	    }
	    if (count > 1) //AAACBBB not AAABBB count must be >1 for true transition
	    {	        
                /* write out the number of single bytes */
                //bitbuf = (bitbuf<<8) | (count - 2);
                //dest[out] = (bitbuf>>numbits);                
                dest[out] = count - 2; //Don't include 1st NOT equal		                			              
                dest[out+1] = c1;
                
                /* Copy single bytes after RLE count */
                memcpy(dest+out+1,srce+idx-count,count-1);
                //memcpy(dest+out+1,input+idx-count,count-1);		               		                
		//cnt = 2; //2-byte pair
		cnt = count; //RLE count + number of single bytes

                pdx = out;
                val = count - 2;
                Printf ("EncodeRLE - RLE actual value: %d\n", val);
                Printf ("EncodeRLE - RLE loop value: %ld\n", count-1);
                Printf ("EncodeRLE - RLE index value: %ld\n", (idx-count)+1); //length=(count-1) //cnt=count
                //Printf ("EncodeRLE - RLE whole byte: %d\n", c1);

                /* reset run */
                out += cnt;
                count = 0;
            }
            if (count == 1) count = 0;            
        }        
        else /* There is No byte run - value is different */
        {            
            Printf ("EncodeRLE - Bytes are NOT equal!!\n");
            count++; /* increment count */
            if (equal == 2 && dest[pdx] <= 128)
            {
                count += equal;
                equal = 1;                
                Printf ("EncodeRLE - Found 2-byte count!! %d\n", c1);
                Printf ("EncodeRLE - Previous index: %ld\n", pdx+1);
                Printf ("EncodeRLE - Previous value: %ld\n", dest[pdx]);     
                count += dest[pdx]+1; //Previous count will be overwritten                                
                out = pdx; //pdx is previous out location
            }
            if ((equal > 1) && (equal <= 128))
            {
                /* write out the number of equal bytes */
                /* rle byte must be encoded as (256 - (equal - 1)) */
                //bitbuf = (bitbuf<<8) | (256 - (equal - 1));
                //dest[out] = (bitbuf>>numbits); //numbits = 0                                              
                dest[out] = 256 - (equal - 1);
                
                //if(depth < 8) //start encode by depth
                //{
                    //numbits += depth;
                    //bitbuf = (bitbuf<<depth) | c1; //load the byte
                    //dest[out+1] = bitbuf<<(8-numbits); //encoded byte
                    //val = dest[out+1];
                    //Printf ("EncodeRLE - RLE encoded value: %d\n", val);
                //}
                //else
                //{
                    dest[out+1] = c1; //unencoded byte
                //}
		cnt = 2; //2-byte pair

                val = 256 - (equal - 1);
                Printf ("EncodeRLE - RLE actual value: %d\n", val);
                Printf ("EncodeRLE - RLE loop value: %d\n", equal);
                Printf ("EncodeRLE - RLE whole byte: %d\n", c1);

                /* reset run */
                out += cnt;
                equal = 1;
            }
            if (count > 0)
		    {
		    Printf ("EncodeRLE - Single count value: %ld\n", count);                
                
				//if((count == 1) && (c1 < 128))	//0x80		          
		        //{ 
		            //dest[out] = c1;		                
		            //cnt = 1; //single byte
                    //Printf ("EncodeRLE - RLE single byte: %d\n", c1);
		        //}
				
            }           
        }
        c1 = c2;	
    }

    /* Finish up last byte */
    if ((count > 1) || (equal > 1))	
    {
        if (count > 1)
        {
            Printf ("EncodeRLE - Finish Encoding!!\n");            
            /* write out the number of single bytes */
            dest[out] = count - 2; //Don't include 1st NOT equal
            /* Copy single bytes after RLE count */
            memcpy(dest+out+1,srce+idx-count,count-1);
            //memcpy(dest+out+1,input+idx-count,count-1);
	    cnt = count; //RLE count + number of single bytes

            val = count - 2;
            Printf ("EncodeRLE - RLE actual value: %d\n", val);
            Printf ("EncodeRLE - RLE loop value: %ld\n", count-1);
            Printf ("EncodeRLE - RLE current index: %ld\n", idx);
            Printf ("EncodeRLE - RLE index value: %ld\n", (idx-count)+1); //length=(count-1) //cnt=count                            
            out += cnt;            
        }
        if (equal > 1)
        {
            /* write out the number of equal bytes */
            /* rle byte must be encoded as (256 - (equal - 1)) */                              
            dest[out] = 256 - (equal - 1);		                			              
            dest[out+1] = c1;		               		                
            cnt = 2; //2-byte pair

            val = 256 - (equal - 1);
            Printf ("EncodeRLE - RLE actual value: %d\n", val);
            Printf ("EncodeRLE - RLE loop value: %d\n", equal);
            Printf ("EncodeRLE - RLE whole byte: %d\n", c1);

            /* reset run */
            out += cnt;
            //equal = 1;
        }
        //out += cnt;
    }
    /* Check Compression */
    if(out > len) /* no RLE */
    {
        out = len;
        *dtype = 0;
        for(i = 0; i < len; ++i)
	    {
            //dest[i]= srce[i];
            dest[i]= input[i];	        
            Printf ("createrle - no RLE byte value: %d\n", dest[i]);
	    }
    }
    else
    *dtype = 1;    
    *dsize = out;

	Printf ("EncodeRLE - total entries: %ld\n", len);
	Printf ("EncodeRLE - compressed size: %ld\n", out);

    //return out; 
    return dest;  
}

/**************************************************************************************************/

//static char *DecodeNew(unsigned char *buffer, long numbytes, long bits, struct IconImage *im, enum DecodeType type, long mode)
static char *DecodeRLE(unsigned char *srce, unsigned long bytes, unsigned long len, struct IconImage *im)
{
    /**************************************/
    /*    FUNCTION BASED ON UNPACKBITS    */
    /**************************************/

    //https://en.wikipedia.org/wiki/PackBits

    
    int i, idx;
    //int i, idx, cnt;
    unsigned char val;
    unsigned char *dest;
    long out, copy, loop;    
    
    idx = 0;
    out = 0;
    copy = 0;
    loop = 0;    

    dest = malloc(len * 2);
    if (!dest) return NULL;

    //Printf ("DecodeRLE - Number of Bitpanes: %ld\n", depth);

    unsigned char input[] = {0xFE, 0xAA, 0x02, 0x80, 0x00, 0x2A, 0xFD, 0xAA, 0x03, 0x80, 0x00, 0x2A, 0x22, 0xF7, 0xAA};       
    // original = 'FE AA 02 80 00 2A FD AA 03 80 00 2A 22 F7 AA'
    // output is: 'AA AA AA 80 00 2A AA AA AA AA 80 00 2A 22 AA AA AA AA AA AA AA AA AA AA'
    
    //unsigned char input[] = {FE AA FE C6 FE E3 FE FF FE 8A FE AA FE 6A FE 29 FE 49 02 56 B1 5A FE 00 11 19 5E 32 38 88 46 24 42 66 52 66 94 7F 89 C2 AD AD F0}
    // output is: 'AA AA AA C6 C6 C6 E3 E3 E3 FF FF FF 8A 8A 8A AA AA AA 6A 6A 6A 29 29 29 49 49 49 56 B1 5A 00 00 00 19 5E 32 38 88 46 24 42 66 52 66 94 7F 89 C2 AD AD F0'

    //unsigned char input[] = {FE AA 05 DF BA 45 EF E7 14 FE FF FE AA FE C6 FE E3 FE 6A FE 8A FE 49 FE 00 FE 29 0E 38 88 46 19 5E 32 24 42 66 52 66 94 7F 89 C2};    
    // output is: 'AA AA AA DF BA 45 EF E7 14 FF FF FF AA AA AA C6 C6 C6 E3 E3 E3 6A 6A 6A 8A 8A 8A 49 49 49 00 00 00 29 29 29 38 88 46 19 5E 32 24 42 66 52 66 94 7F 89 C2'

    //len = 15;
    //bytes = 24;
    while (idx < len && out <= bytes)
    {
        while (!copy && !loop)
        {
            //val = input[idx];
            val = srce[idx]; //load byte
            Printf ("DecodeRLE - RLE actual value: %d\n", val);            

            if (val <= 127)
            {
                (copy = val + 1);
                idx++; //RLE control byte                
                Printf ("DecodeRLE - RLE copy value: %ld\n", copy);                
            }
            else if (val > 128) 
            {
                loop = (256 - val +1 );
                idx++; //RLE control byte
                Printf ("DecodeRLE - RLE loop value: %ld\n", loop);
            }
            else
            continue;
            //Printf ("DecodeRLE - RLE index value: %ld\n", idx);
        }

        while (copy || loop)
        {
            if (loop)
            {
                val = srce[idx]; //load next byte input[idx];
                Printf ("DecodeRLE - RLE byte value: %ld\n", val);

                for (i=0; i<loop; i++)
                {
                    dest[out+i] = val;
                    //im->ImageData[out+i] = val;
                }
                idx++; //RLE loop byte                
                out += loop;
                loop = 0;
            }
            if (copy)
            {
                for (i=0; i<copy; i++)
                {
                    val = srce[idx]; //load single byte input[idx];
                    Printf ("DecodeRLE - RLE byte value: %ld\n", val);                    
                    dest[out+i] = val;
                    //im->ImageData[out+i] = val;
                    idx++; //RLE copy byte
                }
                out += copy;
                copy = 0;
            }
        }
    } //end while idx    
    Printf ("DecodeRLE - RLE out value: %ld\n", out);
    Printf ("DecodeRLE - RLE bytes value: %ld\n", bytes);

    //return out != bytes ? "error decoding iff icon data" : 0;
    return dest;
}

/**************************************************************************************************/

static char *DecodeNew(unsigned char *buffer, long numbytes, long bits, struct IconImage *im, enum DecodeType type, long mode)
{

  /**************************************/
  /*    FUNCTION BASED ON MODIFYICON    */
  /**************************************/
  
  int cop = 0, loop = 0, entries, numbits = 0, mask, curentry = 0;
  unsigned long bitbuf = 0;
  unsigned char byte; 

  Printf ("DecodeNew - Number of Bits: %ld\n", bits); 

  if(!mode) /* no RLE */
    cop = numbytes*8/bits;
  if(type == DECODE_PALETTE)
    entries = im->NumColors*3;
  else
    entries = im->Width*im->Height;

  mask = (1<<bits)-1;
  Printf ("DecodeNew - mask: %d\n", mask);

  if(!cop) Printf ("DecodeNew - !cop = RLE\n");
  else if(cop) Printf ("DecodeNew - cop = no RLE\n");

  while((numbytes || (cop && numbits >= bits)) && (curentry < entries))
  {
    if(!cop)
    {        
        if(numbits < 8)
        {
            bitbuf = (bitbuf<<8)|*(buffer++); --numbytes;
            //byte = (bitbuf<<8)|*(buffer++);
            Printf ("DecodeNew - RLE upper bitbuf byte: %d\n", (unsigned char)bitbuf);
            numbits += 8;
        }
        byte = (bitbuf>>(numbits-8))&0xFF;
        Printf ("DecodeNew - RLE numbits value: %d\n", numbits);
        numbits -= 8;
        if(byte <= 127)
        {
            cop = byte+1;
            Printf ("DecodeNew - RLE cop value: %d\n", cop);
        }
        else if(byte > 128) 
        {
            loop = 256-byte+1;
            Printf ("DecodeNew - RLE loop value: %d\n", loop);
        }
        else                
            continue;
    }
    if(cop) ++loop;

    if(numbits < bits)
    {
      bitbuf = (bitbuf<<8)|*(buffer++); --numbytes;
      numbits += 8;
    }
    byte = (bitbuf>>(numbits-bits))&mask;
    //byte = (bitbuf &0xFF);
    Printf ("DecodeNew - RLE byte value: %d\n", byte);
    while(loop && (curentry < entries))
    {
      if(type == DECODE_PALETTE)
      { /* NOTE: This may not be very fast, but for sure portable */
        switch(curentry%3)
        {
        case 0: im->Palette[curentry/3].Red = byte; break;
        case 1: im->Palette[curentry/3].Green = byte; break;
        case 2: im->Palette[curentry/3].Blue = byte; break;
        }
      }
      else
        im->ImageData[curentry] = byte;
      ++curentry;
      --loop;
    }
    if(cop) --cop;
    numbits -= bits;
  }
  //if(type == DECODE_PALETTE)
      //WriteData("Ram Disk:pal.bin", (unsigned char*)im->Palette, im->NumColors*3);
  //else
      //WriteData("Ram Disk:img.bin", im->ImageData, im->Width*im->Height);  

  return curentry != entries ? "error decoding new icon data" : 0;
}

/**************************************************************************************************/

//static char *DecodeNew(unsigned char *buffer, long numbytes, long bits, struct IconImage *im, enum DecodeType type, long mode)
static char *DecodeBytes(unsigned char *srce, unsigned long bytes, unsigned long depth, struct IconImage *im, unsigned long len)
{
    /*************************************/
    /*    FUNCTION BASED ON DECODERLE    */
    /*************************************/       
    
    int mask;
    long bits;
    int i, idx;    
    unsigned char val;
    //unsigned char *dest;
    long out, copy, loop; 
    unsigned long bitbuf = 0;   
    
    idx = 0;
    out = 0;
    copy = 0;
    loop = 0; 
    bits = 0;   

    //dest = malloc(len * 2);
    //if (!dest) return NULL;    

    Printf ("DecodeBytes - Number of Bitpanes: %ld\n", depth);
    mask = (1<<depth)-1; //numcolors - 1

    unsigned char input[] = {0xFE, 0xAA, 0x02, 0x80, 0x00, 0x2A, 0xFD, 0xAA, 0x03, 0x80, 0x00, 0x2A, 0x22, 0xF7, 0xAA};       
    // original = 'FE AA 02 80 00 2A FD AA 03 80 00 2A 22 F7 AA'
    // output is: 'AA AA AA 80 00 2A AA AA AA AA 80 00 2A 22 AA AA AA AA AA AA AA AA AA AA'
    
    //unsigned char input[] = {FE AA FE C6 FE E3 FE FF FE 8A FE AA FE 6A FE 29 FE 49 02 56 B1 5A FE 00 11 19 5E 32 38 88 46 24 42 66 52 66 94 7F 89 C2 AD AD F0}
    // output is: 'AA AA AA C6 C6 C6 E3 E3 E3 FF FF FF 8A 8A 8A AA AA AA 6A 6A 6A 29 29 29 49 49 49 56 B1 5A 00 00 00 19 5E 32 38 88 46 24 42 66 52 66 94 7F 89 C2 AD AD F0'

    //unsigned char input[] = {FE AA 05 DF BA 45 EF E7 14 FE FF FE AA FE C6 FE E3 FE 6A FE 8A FE 49 FE 00 FE 29 0E 38 88 46 19 5E 32 24 42 66 52 66 94 7F 89 C2};    
    // output is: 'AA AA AA DF BA 45 EF E7 14 FF FF FF AA AA AA C6 C6 C6 E3 E3 E3 6A 6A 6A 8A 8A 8A 49 49 49 00 00 00 29 29 29 38 88 46 19 5E 32 24 42 66 52 66 94 7F 89 C2'

    //len = 15;
    //bytes = 24;
    while (idx < len && out <= bytes)
    {
        while (!copy && !loop)
        {
            //val = input[idx];
            //val = srce[idx]; //load byte
            if(bits < 8)
            {
                bitbuf = (bitbuf<<8)|*(srce++); //load byte input[idx];
                Printf ("DecodeBytes - Uppder bitbuffer value: %d\n", (unsigned char)bitbuf);               
                bits += 8;
            }
            Printf ("DecodeBytes - RLE bits value: %ld\n", bits);
            val = (bitbuf>>(bits-8))&0xFF;  //get byte with mask
            bits -= 8;
            Printf ("DecodeBytes - RLE actual value: %d\n", val);            

            if (val <= 127)
            {
                (copy = val + 1);
                idx++; //RLE control byte                
                Printf ("DecodeBytes - RLE copy value: %ld\n", copy);                
            }
            else if (val > 128) 
            {
                loop = (256 - val +1 );
                idx++; //RLE control byte
                Printf ("DecodeBytes - RLE loop value: %ld\n", loop);
            }
            else
            continue;
            Printf ("DecodeBytes - RLE index value: %ld\n", idx);
        }

        while (copy || loop)
        {
            if (loop)
            {
                //val = srce[idx]; //load next byte
                if(bits < depth)
                {
                    bitbuf = (bitbuf<<8)|*(srce++); //load byte input[idx];
                    Printf ("DecodeBytes - Lower bitbuffer value: %d\n", (unsigned char)bitbuf);               
                    bits += 8;
                }                
                val = (bitbuf>>(bits-depth))&mask; //get byte with mask
                Printf ("DecodeBytes - RLE byte value: %ld\n", val);

                for (i=0; i<loop; i++)
                {
                    //dest[out+i] = val;
                    im->ImageData[out+i] = val;
                }
                idx++; //RLE loop byte 
                bits -= depth;               
                out += loop;                
                loop = 0;
            }
            if (copy)
            {
                for (i=0; i<copy; i++)
                {
                    //val = srce[idx]; //load single byte
                    if(bits < depth)
                    {
                        bitbuf = (bitbuf<<8)|*(srce++); //load byte input[idx]
                        Printf ("DecodeBytes - Lower bitbuffer value: %d\n", (unsigned char)bitbuf);               
                        bits += 8;
                    }                
                    val = (bitbuf>>(bits-depth))&mask; //get byte with mask
                    Printf ("DecodeBytes - RLE byte value: %ld\n", val);

                    //dest[out+i] = val;
                    im->ImageData[out+i] = val;
                    idx++; //RLE copy byte
                    bits -= depth;
                }
                out += copy;
                copy = 0;
            }
        }
    } //end while idx    
    Printf ("DecodeBytes - RLE out value: %ld\n", out);
    Printf ("DecodeBytes - RLE bytes value: %ld\n", bytes);

    return out != bytes ? "error decoding iff icon data" : 0;
    //return dest;
}

/**************************************************************************************************/

//static char *EncodeNew(unsigned long depth, unsigned char *dtype, LONG *dsize, unsigned long size,
    	    	        //unsigned char *src)
static char * createrle(unsigned long depth, unsigned char *dtype, LONG *dsize, unsigned long size,
    	    	        unsigned char *src)
{
	/**************************************/
    /*    FUNCTION BASED ON MODIFYICON    */
    /**************************************/
    
	int i, j, k;    
    unsigned char *buf;
	unsigned char byte;
	unsigned char loop;
	unsigned char loopbytes;
	unsigned char singlebytes;
	unsigned long bitbuf, numbits;
    long ressize, numcopy, numequal;
	//entries = size

    buf = malloc(size * 2);
    if (!buf) return NULL;
  
    numcopy = 0;
    numequal = 1;
    bitbuf = 0;
    numbits = 0;
    ressize = 0;
    k = 0; /* the real output pointer */


    for(i = 1; numequal || numcopy;)
    {
        if(i < size && numequal && (src[i-1] == src[i]))
        {
            ++numequal; ++i; //loop byte
        }
        else if(i < size && numequal*depth <= 16) //not equal
        {
            /* counts the single bytes till the next equal byte */
			numcopy += numequal; //increment numcopy
			numequal = 1; ++i; //reset numequal
        }
        else
        {
            /* care for end case, where it maybe better to join the two */
			//if(i < size && numcopy + numequal <= 128 && (numequal-1)*depth <= 8) ??
            if(i == size && numcopy + numequal <= 128 && (numequal-1)*depth <= 8)
            {
                //last entry?? //counts the single bytes??
				numcopy += numequal; //capture numequal
				numequal = 0; //reset numequal
				Printf ("createrle - RLE last numcopy value: %ld\n", numcopy);
            }

            if(numcopy) //singlebytes
            {
                if((j = numcopy) > 128) j = 128;
                bitbuf = (bitbuf<<8) | (j-1);
                numcopy -= j; //reset numcopy
            }
            else //numequal, loopbytes
            {
                if((j = numequal) > 128) j = 128;
                bitbuf = (bitbuf<<8) | (256-(j-1));                
                byte = bitbuf;
                Printf ("createrle - RLE upper bitbuf byte: %d\n", byte);
                numequal -= j; //reset numequal = 0
                k += j-1;
                j = 1;
            }
            byte = bitbuf;
            Printf ("createrle - RLE bitbuf byte: %d\n", byte);
            buf[ressize++] = (bitbuf >> numbits);
	        byte = (bitbuf >> numbits); //numbits = 0
            Printf ("createrle - RLE actual value: %d\n", byte);
	        if(byte >= 128) loop = (256-byte) + 1; //loop byte
	        else if(byte < 128) loop = byte + 1; //single byte
	        //if(loop) loopbytes += 2; //control byte + byte value
		if(byte > 128)
	        Printf ("createrle - RLE loop value: %d\n", loop);

            while(j--) //process one time for each j
            {
                numbits += depth; //start encode by depth
                bitbuf = (bitbuf<<depth) | src[k++]; //load the byte
                byte = src[k++];
                Printf ("createrle - RLE next byte: %d\n", byte);
                byte = bitbuf;
                Printf ("createrle - RLE lower bitbuf byte: %d\n", byte);
                Printf ("createrle - RLE numbits byte: %d\n", numbits);

                if(numbits >= 8) //for pal numbits = 8
                {
                    numbits -= 8; //reset numbits = 0
                    buf[ressize++] = (bitbuf >> numbits);
		            byte = (bitbuf >> numbits); //numbits = 0
                    //Printf ("createrle - RLE whole byte: %d\n", byte);
                }
            }

            if(i < size && !numcopy && !numequal)
            {
                numequal = 1; ++i; //reset numequal = 1
            }
        }//end else /* care for end case */
    }//end for(k = 0...)

    if(numbits) //finish encode by depth
	{
        buf[ressize++] = bitbuf << (8-numbits);
        byte = bitbuf << (8-numbits);
        Printf ("createrle - next encoded byte: %d\n", byte);
	}

    if(ressize > size) /* no RLE */
    {
        ressize = size;
        *dtype = 0;
        for(i = 0; i < size; ++i)
	    {
            buf[i]= src[i];
	        byte = buf[i];
            Printf ("createrle - no RLE byte value: %d\n", byte);
	    }
    }
    else
    *dtype = 1;    
    *dsize = ressize;

	Printf ("createrle - total entries: %ld\n", size);
	Printf ("createrle - RLE loopbytes: %d\n", loopbytes);
  
    return buf;
}

/**************************************************************************************************/

BOOL ReadIconImage_IFF(char *srcename, LONG iffoffset)
{
    ULONG id;
	char *error;	    
    LONG width = 0;
    LONG height = 0;
    LONG cksize = 0; //chunk size
    LONG icsize = 0; //form size
    LONG imSize = 0; //file size
    LONG offset = 0;     
    	        
    BOOL retval = TRUE;	
	unsigned char *buffer;
    
	LONG numimages = 0, numface = 0;
	LONG transparency, numcolors, icflags;  
	LONG imgfmt, palfmt, numimg, numpal, depth;   


    /* Get total icon file size */
	imSize = GetFileSize(srcename);
	Printf ("File Size: %ld\n", imSize);

    /* Get byte array of IFF data */
    //buffer = (unsigned char *)AllocVec(imSize-iffoffset, MEMF_ANY);
    buffer = AllocMem(imSize-iffoffset, MEMF_CLEAR);
    //FreeMem(buffer, imSize-iffoffset);

    /* Open Icon File 0xE310 to Read */    
    BPTR fh = Open(srcename, MODE_OLDFILE);
    if (fh)
    {
        /* First Offset */
        offset += iffoffset;

        /* Beginning of OS3.5 IFF Data */
		Seek(fh,offset,OFFSET_BEGINNING);
		Read(fh, buffer, imSize);

		//0x464F524D = 'FORM' //0x49434F4E = 'ICON'
		//if(GetUint32(buffer) != ID_FORM || GetUint32(buffer+8) != ID_ICON)
		if(GetUint32(buffer) != 0x464F524D || GetUint32(buffer+8) != 0x49434F4E)		
		    retval = FALSE;

		icsize = GetUint32(buffer+4)-4;
		//Printf ("Icon Form Size: %ld\n", icsize);
		buffer += 12;

		LONG ilbmsize = 0;
		LONG formsize = 0;

        struct IconImage *im;
        struct IconImageCollection *ico;

        im = (struct IconImage *)AllocMem(sizeof(struct IconImage), MEMF_CLEAR);
        ico = (struct IconImageCollection *)AllocMem(sizeof(struct IconImageCollection), MEMF_CLEAR);
        /* Don't free mem for these because we need to pass them to other functions */
        //FreeMem(im, sizeof(struct IconImage));
        //FreeMem(ico, sizeof(struct IconImageCollection));

        /* Buffers for encoded img data & pal data */
        LONG palsize, imgsize;
        UBYTE palpacked, imgpacked;
        //LONG ilbm1_cmapsize = 0; //palsize
        //LONG ilbm2_cmapsize = 0; //palsize
        //LONG ilbm1_planarsize = 0; //imagesize
        //LONG ilbm2_planarsize = 0; //imagesize
        //unsigned char *ilbm1_planardata, *ilbm1_planarcmap;
	//unsigned char *ilbm2_planardata, *ilbm2_planarcmap;
	unsigned char *ilbm_planardata, *ilbm_planarcmap;

		while(icsize > 0)
		{
			id = GetUint32(buffer);
            cksize = GetUint32(buffer+4);
            buffer += 8;
			switch(id)
			{
				case 0x46414345: /* FACE */ //ID_FACE
				    Printf ("Found FACE chunk\n");
					Printf ("Chunk Size: %ld\n", cksize);				    
					//FACE chunk must be at least 6 bytes
					if(!numface) //&& cksize >= 6)
					{
						++numface;						
				        width = buffer[0]+1;
                        height = buffer[1]+1;
						Printf ("FACE Dimensions: %ld, %ld\n", width, height);						
					}
				    break;
                case 0x494D4147: /* IMAG */ //ID_IMAG
				    Printf ("Found IMAG chunk\n");
					Printf ("Chunk Size: %ld\n", cksize);
					if(cksize >= 10)
					{                       
                        /* IMAG chunk data */                        
                        im = ico->NewImage+numimages;
			            im->Width = width;
                        im->Height = height;
						transparency = buffer[0];						
                        im->NumColors = buffer[1]+1;
                        im->Transparency = transparency;
                        icflags = buffer[2];
                        imgfmt = buffer[3];
                        palfmt = buffer[4];
                        depth = buffer[5];
                        im->Depth = depth;
                        numimg = GetUint16(buffer+6)+1;
                        numpal = GetUint16(buffer+8)+1; //Size of pal data
						numcolors = 1<<depth; //No. colors by depth						 
						
			            Printf ("Dimensions: %ld, %ld, %ld\n", width, height, depth);
			            Printf ("Transparency: %ld\n", transparency);
			            Printf ("Num Colors: %ld\n", numcolors);
			            Printf ("IC Flags: %ld\n", icflags);
			            Printf ("Image Format: %ld\n", imgfmt);
			            Printf ("Palette Format: %ld\n", palfmt);
			            Printf ("Image Depth: %ld\n", depth);
			            Printf ("Image Size: %ld\n", numimg);
			            Printf ("Palette Size: %ld\n", numpal);
			            Printf ("Number of Colors: %ld\n", im->NumColors*3);						

						if(imgfmt <= 1 && palfmt <= 1) //It's zero or one
						{							
							//Printf ("Decoding Image...\n");
			    //static char *DecodeNew(unsigned char *buffer, long numbytes, long bits, struct IconImage *im, enum DecodeType type, long mode)				
                            
                            //error = DecodeNew(buffer+10, numimg, imgfmt ? depth : 8, im, DECODE_IMAGE, imgfmt);							
                            //if(error) Printf ("Error: %s\n", error);
                            
                            //static char *DecodeBytes(unsigned char *srce, unsigned long bytes, unsigned long depth, struct IconImage *im, unsigned long len)                            
                            //unsigned char *imgdata = DecodeBytes(buffer+10, width*height, imgfmt ? depth : 8, im, numimg);
                            
                            error = DecodeBytes(buffer+10, width*height, imgfmt ? depth : 8, im, numimg);
                            if(error) Printf ("Error: %s\n", error);   
                            
                            ilbm_planardata = createrle(depth, &imgpacked, &imgsize,
		                        width * height, im->ImageData);
                            //ilbm1_planarsize = imgsize;
		                    Printf("IMAG planar size: %ld\n", imgsize);
                            //WriteData("Ram Disk:img2.bin", ilbm1_planardata, imgsize);  

                            if(icflags & 2)
                            {
                                //Printf ("Decoding Palette...\n");
                                //error = DecodeNew(buffer+10+numimg, numpal, 8, im, DECODE_PALETTE, palfmt);
                                //if(error) Printf ("Error: %s\n", error);
                                
                                //static char *DecodeRLE(unsigned char *srce, unsigned long bytes, unsigned long len, struct IconImage *im)
                                //error = DecodeRLE(buffer+10+numimg, width*height, numpal, im);
                                unsigned char *paldata = DecodeRLE(buffer+10+numimg, im->NumColors*3, numpal, im);
                                //if(error) Printf ("Error: %s\n", error);
                                SetPalette(im, (struct IconPalette*)paldata, im->NumColors);
                                
                                WriteData("Ram Disk:pal.bin", (unsigned char*)im->Palette, im->NumColors*3); 


                                //ilbm_planarcmap = createrle(8, &palpacked, &palsize, im->NumColors * 3,
			                        //(unsigned char *)im->Palette);
			                        
	                            //ilbm1_cmapsize = palsize;	                            
                                //static char *EncodeRLE(8, &palpacked, &palsize, int len, unsigned char *srce);
                                //im->NumColors * 3 //24
                                
                                ilbm_planarcmap = EncodeRLE(8, &palpacked, &palsize, im->NumColors * 3,
			                        (unsigned char *)im->Palette);			                        
		                        
                                //WriteData("Ram Disk:pal2.bin", ilbm1_planarcmap, palsize);                               
                            }
                            else if(numimages)
                            {
                                //Printf ("Copying Palette...\n");
                                im->NumColors = ico->NewImage[0].NumColors; //IconImageCollection
                                SetPalette(im, ico->NewImage[0].Palette, im->NumColors);
                            }                                
						}						
					}
					//Error! IMAG chunk must be at least 10 bytes if present.
					
                    if(numimages == 0)
					{                        
                        //Printf ("Saving ILBM File...\n");                               
                        SaveImage("Ram Disk:img1.iff", im);
                        WriteData("Ram Disk:img1.bin", ilbm_planardata, imgsize);
                        WriteData("Ram Disk:pal1.bin", ilbm_planarcmap, palsize);  
					}
					if(numimages == 1)
					{						
                        //Printf ("Saving ILBM File...\n");
                        SaveImage("Ram Disk:img2.iff", im);
                        WriteData("Ram Disk:img2.bin", ilbm_planardata, imgsize);
                        WriteData("Ram Disk:pal2.bin", ilbm_planarcmap, palsize);  
					}
		
					++numimages; //Number of images present in icon file
					break;
				default:
				    break;
			}
			if(cksize & 1) ++cksize;
            buffer += cksize;
            icsize -= 8+cksize;
		}
		if(icsize < 0)
            retval = FALSE;
    }
    Close(fh);
	Printf ("Number of Images: %ld\n", numimages);

    if(buffer) FreeMem(buffer, imSize-iffoffset);

    return retval;
}

/**************************************************************************************************/

int getarguments(int argc, char **argv)
{
    int retvalue = 0;

    if (argc == 0)
        retvalue = 0;
	if (argc == 1)
        retvalue = 1;
	if (argc == 2)
        retvalue = 2;

    return retvalue;
}

/**************************************************************************************************/

int main(int argc, char **argv)
{	
    char *type = NULL;
    BOOL error = FALSE;
    LONG iffoffset = 0;

	int action = getarguments(argc, argv);
    Printf ("Action Value: %d\n", action);
	
    /* CLI Arguments & Switches */
    //iconsplit iconfile.info returns 1

    if(action == 1 || action == 2)
    //if(argc == 1 || argc == 2)
    {
        srcename = argv[1]; /* source icon */

        /* Verify 'E310' Icon */
        type = ReadIcon(srcename);        

        if(stricmp(type, "iff")==0)
        {
            Printf ("Found Classic Icon\n");
            /* Find Beginning of OS3.5 IFF Data */
	        iffoffset = Get_IFFOffset(srcename);            
            Printf ("\nReturned IFF Offset: %ld\n", iffoffset);

            /* Find IFF IMAG data & Save ILBM */            		
		    error = ReadIconImage_IFF(srcename, iffoffset);
        }
	    else
	    {
            Printf ("Not a valid Classic Icon\n");
	    }
    }
    else
	{
        Printf ("Not a valid argument\n");
	}

	return 0;
}

/**************************************************************************************************/